/*****************************************************************************
 * osiofile.c: DVD OS dependant reading library.
 * Borrow from VideoLAN
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "defs.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#if defined( WIN32 )
#   include <windows.h>
#   include <io.h>                                                 /* read() */
#else
#   include <unistd.h>
#endif

#include "dvd_ioctl.h"
#include "input_dvd.h"
#include "osfileio.h"

#define DVDCSS_BLOCK_SIZE DVD_LB_SIZE

/*****************************************************************************
 * Local prototypes, win32 specific
 *****************************************************************************/
#if defined( WIN32 )
static int   _win32_dvdcss_aopen      ( char c_drive);
static int   _win32_dvdcss_aclose     ( int i_fd );
static s64   _win32_dvdcss_aseek      ( int i_fd, s64 offset);
static int   _win32_dvdcss_aread      ( int i_fd, void *p_data, int i_blocks );
static int _win32_dvdcss_aread_normal ( int i_fd, void *p_data, int i_size );
#endif

int dvd_os_open  ( const char *device_name)
{
 int i_fd;
#if defined( WIN32 )
    if( WIN2K )
    {
        char psz_dvd[16];
        sprintf( psz_dvd,"\\\\.\\%c:", device_name[0] );
        (HANDLE)i_fd =
                CreateFile( psz_dvd, GENERIC_READ | GENERIC_WRITE,
                                FILE_SHARE_READ | FILE_SHARE_WRITE,
                                NULL, OPEN_EXISTING, 0, NULL );
        if( (HANDLE) i_fd == INVALID_HANDLE_VALUE )
        {
           return -1;
        }
    }
    else
    {
        i_fd = _win32_dvdcss_aopen(device_name[0]);
    }
#else
    i_fd = open( device_name, O_RDONLY | O_NONBLOCK );
#endif
    return i_fd;
}

int   dvd_os_close ( int handle)
{
#if defined( WIN32 )
    if( WIN2K )
    {
        CloseHandle( (HANDLE) handle);
    }
    else
    {
        _win32_dvdcss_aclose(handle);
    }

#else
    close( handle );
#endif
    return 0;
}

s64 dvd_os_lseek ( int handle, s64 offset )
{
#if defined( WIN32 )
    if( WIN2K )
    {
        DWORD lopart;
        DWORD hipart;

        lopart = (DWORD) (offset & 0xffffffff);
        hipart = (DWORD) ((u64) offset >> 32);

        lopart = SetFilePointer( (HANDLE) handle,
                                          lopart,
                                          &hipart, FILE_BEGIN );
        if( (lopart == 0xffffffff)
            && GetLastError() != NO_ERROR)
        {
            return (s64) -1;
        }
        return ((u64)lopart + ((u64)hipart << 32));
    }
    else
    {
        return ( _win32_dvdcss_aseek(handle, offset) );
    }

#else
    return lseek( handle,offset,SEEK_SET );
#endif

}

int dvd_os_read  ( int handle, void *p_buffer, int i_size )
{
#if defined( WIN32 )
    int i_bytes;
    if( WIN2K )
    {
        if( !ReadFile( (HANDLE) handle, p_buffer,
                        i_size,
                       (LPDWORD)&i_bytes, NULL ) )
        {
            return -1;
        }
    }
    else
    {
        i_bytes = _win32_dvdcss_aread_normal( handle, p_buffer, i_size);
    }

    return i_bytes;
#else
    return read (handle, p_buffer ,i_size);
#endif

}


#if defined( WIN32 )

/*****************************************************************************
 * _win32_dvdcss_aopen: open dvd drive (load aspi and init w32_aspidev
 *                      structure)
 *****************************************************************************/
static int _win32_dvdcss_aopen( char c_drive)
{
    HMODULE hASPI;
    DWORD dwSupportInfo;
    struct w32_aspidev *fd;
    int i, j, i_hostadapters;
    long (*lpGetSupport)( void );
    long (*lpSendCommand)( void* );
     
    hASPI = LoadLibrary( "wnaspi32.dll" );
    if( hASPI == NULL )
    {
        printf("unable to load wnaspi32.dll\n" );
        return -1;
    }

    (FARPROC) lpGetSupport = GetProcAddress( hASPI, "GetASPI32SupportInfo" );
    (FARPROC) lpSendCommand = GetProcAddress( hASPI, "SendASPI32Command" );
 
    if(lpGetSupport == NULL || lpSendCommand == NULL )
    {
        printf( "unable to get aspi function pointers\n" );
        FreeLibrary( hASPI );
        return -1;
    }

    dwSupportInfo = lpGetSupport();

    if( HIBYTE( LOWORD ( dwSupportInfo ) ) == SS_NO_ADAPTERS )
    {
        printf("no host adapters found (aspi)\n" );
        FreeLibrary( hASPI );
        return -1;
    }

    if( HIBYTE( LOWORD ( dwSupportInfo ) ) != SS_COMP )
    {
        printf( "unable to initalize aspi layer\n" );
        FreeLibrary( hASPI );
        return -1;
    }

    i_hostadapters = LOBYTE( LOWORD( dwSupportInfo ) );
    if( i_hostadapters == 0 )
    {
        FreeLibrary( hASPI );
        return -1;
    }

    fd = malloc( sizeof( struct w32_aspidev ) );
    if( fd == NULL )
    {
        FreeLibrary( hASPI );
        return -1;
    }

    fd->i_blocks = 0; fd->offset = 0;
    fd->hASPI = (long) hASPI;
    fd->lpSendCommand = lpSendCommand;

    c_drive = c_drive > 'Z' ? c_drive - 'a' : c_drive - 'A';

    for( i = 0; i < i_hostadapters; i++ )
    {
        for( j = 0; j < 15; j++ )
        {
            struct SRB_GetDiskInfo srbDiskInfo;

            srbDiskInfo.SRB_Cmd         = SC_GET_DISK_INFO;
            srbDiskInfo.SRB_HaId        = i;
            srbDiskInfo.SRB_Flags       = 0;
            srbDiskInfo.SRB_Hdr_Rsvd    = 0;
            srbDiskInfo.SRB_Target      = j;
            srbDiskInfo.SRB_Lun         = 0;

            lpSendCommand( (void*) &srbDiskInfo );

            if( (srbDiskInfo.SRB_Status == SS_COMP) &&
                (srbDiskInfo.SRB_Int13HDriveInfo == c_drive) )
            {
                fd->i_sid = MAKEWORD( i, j );
                return (int) fd;
            }
        }
    }

    free( (void*) fd );
    FreeLibrary( hASPI );
    printf( "unable to get haid and target (aspi) \n" );
    return( -1 );        
}

/*****************************************************************************
 * _win32_dvdcss_aclose: close dvd drive (unload aspi and free w32_aspidev
 *                       structure)
 *****************************************************************************/
static int _win32_dvdcss_aclose( int i_fd )
{
    struct w32_aspidev *fd = (struct w32_aspidev *) i_fd;

    FreeLibrary( (HMODULE) fd->hASPI );
    free( (void*) i_fd );

    return 0;
}

/*****************************************************************************
 * _win32_dvdcss_aseek: aspi version of _dvdcss_seek
 * 
 * returns the number of blocks read.
 *****************************************************************************/

static s64 _win32_dvdcss_aseek( int i_fd, s64 offset )
{
    int i_old_blocks;
    char sz_buf[ DVDCSS_BLOCK_SIZE ];
    struct w32_aspidev *fd = (struct w32_aspidev *) i_fd;
    
    i_old_blocks = fd->i_blocks;
    fd->i_blocks = (int) (offset / DVDCSS_BLOCK_SIZE);

    if( _win32_dvdcss_aread( i_fd, sz_buf, 1 ) == -1 )
    {
        fd->i_blocks = i_old_blocks;
        return (s64) -1;
    }

     fd->offset = offset;
    (fd->i_blocks)--;

    return offset;
}
/*****************************************************************************
 * _win32_dvdcss_aread_normal: aspi normal version of _dvdcss_read
 *
 * returns the number of byte read.
 *****************************************************************************/
static int _win32_dvdcss_aread_normal( int i_fd, void *p_data, int i_size )
{
 struct w32_aspidev *fd = (struct w32_aspidev *) i_fd;
 char sz_buf[ DVDCSS_BLOCK_SIZE ];
 s64 next_offset;
 int next_blocks;
 int size_data;
 int total_size_read;
 char *p_buf = (char *) p_data;

   if (i_size<=0) return -1;

   next_offset =  fd->offset + (s64) i_size;
   next_blocks = (int) (next_offset / DVDCSS_BLOCK_SIZE);
   size_data = (int) (fd->offset %  DVDCSS_BLOCK_SIZE);
  
   if (size_data==0 && i_size>=DVDCSS_BLOCK_SIZE)
   {
     total_size_read = 0;
   }
   else
   { 
     if( _win32_dvdcss_aread( i_fd, sz_buf, 1 ) == -1 )
	 {
        return -1;
     }
     total_size_read = DVDCSS_BLOCK_SIZE - size_data;
     memcpy(p_buf,sz_buf + size_data,total_size_read);
     p_buf += total_size_read;
     i_size -= total_size_read;
   }

   size_data = i_size / DVDCSS_BLOCK_SIZE;
   if (size_data>0)
   {
     if( _win32_dvdcss_aread( i_fd, p_buf, size_data ) == -1 )
       {
        return -1;
       }
     size_data *= DVDCSS_BLOCK_SIZE;
     p_buf  +=  size_data;
     i_size  -=  size_data;
     total_size_read += size_data;

     if (i_size>0)
     {
        if( _win32_dvdcss_aread( i_fd, sz_buf, 1 ) == -1 )
        {
          return -1;
        }
        memcpy(p_buf,sz_buf,i_size);
        total_size_read += i_size;
     }
   }
   fd->offset = next_offset;
   fd->i_blocks = next_blocks;
   return total_size_read;
}

/*****************************************************************************
 * _win32_dvdcss_aread: aspi version of _dvdcss_read
 *
 * returns the number of blocks read.
 *****************************************************************************/
static int _win32_dvdcss_aread( int i_fd, void *p_data, int i_blocks )
{
    HANDLE hEvent;
    struct SRB_ExecSCSICmd ssc;
    struct w32_aspidev *fd = (struct w32_aspidev *) i_fd;

    /* Create the transfer completion event */
    hEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
    if( hEvent == NULL )
    {
        return -1;
    }

    memset( &ssc, 0, sizeof( ssc ) );

    ssc.SRB_Cmd         = SC_EXEC_SCSI_CMD;
    ssc.SRB_Flags       = SRB_DIR_IN | SRB_EVENT_NOTIFY;
    ssc.SRB_HaId        = LOBYTE( fd->i_sid );
    ssc.SRB_Target      = HIBYTE( fd->i_sid );
    ssc.SRB_SenseLen    = SENSE_LEN;
    
    ssc.SRB_PostProc = (LPVOID) hEvent;
    ssc.SRB_BufPointer  = p_data;
    ssc.SRB_CDBLen      = 12;
    
    ssc.CDBByte[0]      = 0xA8; /* RAW */
    ssc.CDBByte[2]      = (UCHAR) (fd->i_blocks >> 24);
    ssc.CDBByte[3]      = (UCHAR) ((fd->i_blocks >> 16) & 0xff);
    ssc.CDBByte[4]      = (UCHAR) ((fd->i_blocks >> 8) & 0xff);
    ssc.CDBByte[5]      = (UCHAR) ((fd->i_blocks) & 0xff);
    
    /* We have to break down the reads into 64kb pieces (ASPI restriction) */
    if( i_blocks > 32 )
    {
        ssc.SRB_BufLen = 32 * DVDCSS_BLOCK_SIZE;
        ssc.CDBByte[9] = 32;
        fd->i_blocks  += 32;

        /* Initiate transfer */  
        ResetEvent( hEvent );
        fd->lpSendCommand( (void*) &ssc );

        /* transfer the next 64kb (_win32_dvdcss_aread is called recursively)
         * We need to check the status of the read on return */
        if( _win32_dvdcss_aread( i_fd, (u8*) p_data + 32 * DVDCSS_BLOCK_SIZE,
                                 i_blocks - 32) < 0 )
        {
            return -1;
        }
    }
    else
    {
        /* This is the last transfer */
        ssc.SRB_BufLen   = i_blocks * DVDCSS_BLOCK_SIZE;
        ssc.CDBByte[9]   = (UCHAR) i_blocks;
        fd->i_blocks += i_blocks;

        /* Initiate transfer */  
        ResetEvent( hEvent );
        fd->lpSendCommand( (void*) &ssc );

    }

    /* If the command has still not been processed, wait until it's finished */
    if( ssc.SRB_Status == SS_PENDING )
    {
        WaitForSingleObject( hEvent, INFINITE );
    }
    CloseHandle( hEvent );

    /* check that the transfer went as planned */
    if( ssc.SRB_Status != SS_COMP )
    {
      return -1;
    }

    return i_blocks;
}

#endif

